"""Tests for the process function."""
import json
import os
import unittest

import pandas as pd

from tsuniverse.process import process


class TestProcess(unittest.TestCase):

    def setUp(self):
        self.dir = os.path.dirname(__file__)

    def test_process(self):
        df = pd.read_parquet(os.path.join(self.dir, "universe.parquet"))
        features = list(process(df, [df.columns.values.tolist()[0]], 30))
        #with open(os.path.join(self.dir, "expected.json"), "w") as handle:
        #    json.dump(features, handle)
        with open(os.path.join(self.dir, "expected.json")) as handle:
            expected_features = json.load(handle)
            self.assertListEqual(features, expected_features)



"""A list of transforms."""

from .transform import Transform
from .transform_velocity import velocity_transform

TRANSFORMS = {
    Transform.NONE: lambda x: x,
    Transform.VELOCITY: velocity_transform,
}



"""The base class for a feature."""

from typing import NotRequired, TypedDict


class Feature(TypedDict):
    """A description of a feature to use."""

    predictor: str
    predictor_transform: str
    predictand: str
    predictand_transform: str
    lag: NotRequired[float]
    correlation: NotRequired[float]



"""The main process function."""

from multiprocessing import Pool
from typing import Iterator

import pandas as pd

from .feature import Feature
from .pearson_process import pearson_process
from .transforms import TRANSFORMS


def process(
    df: pd.DataFrame,
    predictands: list[str],
    max_window: int,
    max_process_features: int = 10,
) -> Iterator[list[Feature]]:
    """Process the dataframe for tsuniverse features."""
    with Pool() as p:
        for predictand in predictands:
            for transform_name in TRANSFORMS:
                features = sorted(
                    list(
                        pearson_process(df, predictand, max_window, p, transform_name)
                    ),
                    key=lambda x: abs(x["correlation"] if "correlation" in x else 0.0),
                    reverse=True,
                )[:max_process_features]
                yield features



"""The pearson process function."""

from typing import Any, Iterator

import numpy as np
import pandas as pd

from .feature import Feature
from .transform import Transform
from .transforms import TRANSFORMS

_PEARSON_CACHE: dict[str, Feature] = {}


def pearson_correlation_positive_lags(
    x: pd.Series,
    y: pd.Series,
    max_window: int,
    x_transform: Transform,
    y_transform: Transform,
) -> Feature:
    """Calculate the best pearson correlation for the 2 series within a lag window"""
    x = TRANSFORMS[x_transform](x)
    y = TRANSFORMS[y_transform](y)
    corrs = []
    lags = range(1, max_window + 1)

    for lag in lags:
        y_shifted = y.shift(lag)
        valid_idx = x.index.intersection(y_shifted.index)  # type: ignore
        corr = x.loc[valid_idx].corr(y_shifted.loc[valid_idx])
        corrs.append(corr)

    best_idx = np.argmax(np.abs(corrs))
    best_lag = lags[best_idx]
    best_corr = corrs[best_idx]
    if np.isnan(best_corr):
        best_corr = 0.0

    return {
        "predictor": str(y.name),
        "predictor_transform": y_transform,
        "predictand": str(x.name),
        "predictand_transform": x_transform,
        "lag": float(best_lag),
        "correlation": float(best_corr),
    }


def pearson_process(
    df: pd.DataFrame,
    predictand: str,
    max_window: int,
    pool: Any,
    predictand_transform: str,
) -> Iterator[Feature]:
    """Process the dataframe for tsuniverse features."""
    predictors = df.columns.values.tolist()
    cached_predictors = []
    for predictor in predictors:
        for transform in TRANSFORMS:
            key = "_".join(
                sorted([predictor, transform, predictand, predictand_transform])
            )
            feature = _PEARSON_CACHE.get(key)
            if feature is not None:
                yield feature
                cached_predictors.append(predictor)
    for transform in TRANSFORMS:
        for feature in pool.starmap(
            pearson_correlation_positive_lags,
            [
                (df[x], df[predictand], max_window, transform, predictand_transform)
                for x in df.columns.values.tolist()
                if x != predictand and x not in cached_predictors
            ],
        ):
            if feature is None:
                continue
            key = "_".join(
                sorted(
                    [
                        feature["predictor"],
                        transform,
                        feature["predictand"],
                        predictand_transform,
                    ]
                )
            )
            _PEARSON_CACHE[key] = feature
            yield feature
